/*
 *  mk_movi_image.c
 *
 *  Make movinand image with multiple partitions to download onto a movinand
 *
 *  Author : jt77.jang@samsung.com
 *
 *  @ 2010. 6. 24  : first release
 */
#define _GNU_SOURCE
#define _LARGEFILE64_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

#include "partition.h"

#define DEBUG 1

#define VERSION_STRING	"1.0.0"

#define FAIL	 0
#define SUCCESS  1


typedef unsigned int	DWORD;
typedef unsigned char	BYTE;

struct partition_table {
	BYTE partno;         /* partition number */
	DWORD attr;			 /* attribute */
	DWORD partsize;      /* unit : bytes */
	char szFileName[128];
};

/*
					            ==  PARTITION HEADER structure (512 bytes) ==

              4                       4                      4                       4
 +-----------------------+-----------------------+-----------------------+-----------------------+
 |                                           FAT32HEAD                                           |
 +-----------------------+-----------------------+-----------------------+-----------------------+
 |    1st part size      |      not used now     | 1st part start sector |   2nd part size       |
 +-----------------------+-----------------------+-----------------------+-----------------------+
 |     not used now      | 2nd part start sector |     3rd part size     |      not used now     |
 +-----------------------+-----------------------+-----------------------+-----------------------+
 | 3rd part start sector |      4th part size    |      not used now     | 4th part start sector |
 +-----------------------+-----------------------+-----------------------+-----------------------+
 |                                              ...                                              |
 +-----------------------+-----------------------+-----------------------+-----------------------+
 |                       |                       |                       |                       |
 +-----------------------+-----------------------+-----------------------+-----------------------+

*/

typedef struct
{
	BYTE  FAT32HEAD[16];      /* "FATHEAD" identifier */

	DWORD FirstPartImg_Size;  /* Partition image size (bytes unit) */
	DWORD FirstPart_ChkSum;   /* partition checksum data (not used now) */
	DWORD FirstPart_Reserved; /* Start sector of the partition (sector unit) */

	DWORD SecondPart_Size;
	DWORD SecondPart_ChkSum;
	DWORD SecondPart_Reserved;

	DWORD ThirdPart_Size;
	DWORD ThirdPart_ChkSum;
	DWORD ThirdPart_Reserved;

	DWORD FourthPart_Size;
	DWORD FourthPart_ChkSum;
	DWORD FourthPart_Reserved;

	DWORD FifthPart_Size;
	DWORD FifthPart_ChkSum;
	DWORD FifthPart_Reserved;

	DWORD SixthPart_Size;
	DWORD SixthPart_ChkSum;
	DWORD SixthPart_Reserved;

	DWORD SeventhPart_Size;
	DWORD SeventhPart_ChkSum;
	DWORD SeventhPart_Reserved;

	DWORD EightthPart_Size;
	DWORD EightthPart_ChkSum;
	DWORD EightthPart_Reserved;

	/* ... */

	BYTE  reserved[400];	  /* 0x0 */
} PARTITION_HEADER ;

#define PART_INFO_BASE_OFFSET				16
#define PART_INFO_EACH_PART_OFFSET			12
#define PART_INFO_PART_SIZE_OFFSET			0
#define PART_INFO_PART_CHECKSUM_OFFSET		4
#define PART_INFO_START_SECTOR_OFFSET		8


/* partition table entry of MSDOS partition */
struct partition {
	BYTE boot_ind;         /* 0x80 - active */
	BYTE head;             /* starting head */
	BYTE sector;           /* starting sector */
	BYTE cyl;              /* starting cylinder */
	BYTE sys_ind;          /* What partition type */
	BYTE end_head;         /* end head */
	BYTE end_sector;       /* end sector */
	BYTE end_cyl;          /* end cylinder */
	BYTE start4[4];        /* starting sector counting from 0 */
	BYTE size4[4];         /* nr of sectors in partition */
} ATTRIBUTE_PACKED;

#define MBR_RESERVED_SIZE        32768
#define MBR_RESERVED_SECTORS     (MBR_RESERVED_SIZE >> DEFAULT_SECTOR_BITS)

#define DEFAULT_SECTOR_SIZE     512
#define DEFAULT_SECTOR_BITS     9

#define MAX_SECTOR_SIZE		2048

#define LINUX_NATIVE		0x83
#define MS_EX_PART		0x05

#define BUFFER_SIZE		8192
#define BLK_SIZE		262144

#define DOWNLOAD_UNIT		30 * 1024 * 1024 /* 30 MB */
#define FBR_SIZE		512

#define PARTITION_TABLE_OFFSET	446

#define MAX_PARTITION_NUM	12
#define PARTITION_NUM		9


/*
 * Make single image from each partition image
 */
int make_single_image(int image_fd, const char *filename, off64_t offset)
{
	int fd;
	int written;
	char buffer[BUFFER_SIZE];
	int read_size;
	struct stat st;
	unsigned int file_size = 0;
	int total_size = 0;

	if (stat(filename, &st) < 0) {
		perror("stat failed");
		return -1;
	}
	file_size = st.st_size;

	if (DEBUG)
		printf("file_size(byte) : %u\n", file_size);

	printf("open : %s\n", filename);
	if ((fd = open(filename, (O_RDONLY))) < 0) {
		perror("open");
		return -1;
	}

	if (lseek64(image_fd, offset, SEEK_SET) < 0) {
		perror("lseek");
		close(fd);
		return -1;
	}

	printf("read : %s\n", filename);

	while (1) {
		if((read_size = read(fd, buffer, BUFFER_SIZE)) < 0) {
			perror("read");
			close(fd);
			return -1;
		}

		if(!read_size)
			break;

		written = write(image_fd, buffer, read_size);
		if (written < 0) {
			perror("write failed");
			close(fd);
			return -1;
		}
		total_size += written;
	}

	close(fd);

	if (DEBUG)
		printf("### file_size : %u bytes\n", file_size);

	if (total_size != file_size) {
		printf("make_single_image() error\n");
		return -1;
	}

	return file_size;
}

/*
 * Make each partition image with MBR
 *  - create temporary partition image
 *      {partno}.mbr == MBR + PARTITION_IMAGE + padding data for sector alignment
 */
int make_image_with_MBR(PARTITION_HEADER * part_header, unsigned int movi_total_sectors, struct partition_table * pt, int partno)
{
	int i;
	int input_fd = -1, output_fd = -1;
	int written;
	char *mbr_image = NULL;
	struct partition part;
	unsigned int file_size = 0;
	char buffer[DEFAULT_SECTOR_SIZE];
	char temp_buf[256];
	unsigned int start_sector, sector_size;
	unsigned int next_start_sector = 0; /* real position of next partition */
	unsigned int offset = PARTITION_TABLE_OFFSET; /* position of partition table */
	int ret = SUCCESS;

	printf("========================================================\n");
	printf(" Make each partition image with MBR(Master Boot Record) \n");
	printf("========================================================\n");

	if (partno > MAX_PARTITION_NUM || partno < 1) {
		printf("the number of partitions is error : %d\n", partno);
		printf("max : %d, min : 1\n", MAX_PARTITION_NUM);
		return -1;
	}

	mbr_image = malloc(MBR_RESERVED_SIZE);
	if (mbr_image == NULL)
	{
		fprintf(stderr, "Unable to allocate memory.\n");
		return -1;
	}

	for (i = 0; i < partno; i++) {
		file_size = 0;
		memset(mbr_image, 0x00, MBR_RESERVED_SIZE);

		next_start_sector = pt[i].partsize >> DEFAULT_SECTOR_BITS;

		if (DEBUG) {
			printf("pt[i].partsize    : 0x%x\n", pt[i].partsize);
			printf("next_start_sector : 0x%x\n", next_start_sector);
		}

		memcpy((unsigned char *)part_header + PART_INFO_BASE_OFFSET + PART_INFO_START_SECTOR_OFFSET + i * PART_INFO_EACH_PART_OFFSET,
				&next_start_sector, sizeof(next_start_sector));

		/* init offset values */
		offset = PARTITION_TABLE_OFFSET;

		sprintf(temp_buf, "%d.mbr", i);

		printf("open : %s\n", temp_buf);
		if ((output_fd = open(temp_buf, O_CREAT | O_RDWR | O_TRUNC, 0666)) < 0) {
			perror("open");
			ret = FAIL;
			goto mbr_out;
		}

		#if 0 /* CHS address hasn't any meaning on most systems */
		/* starting CHS address */
		part.head = 0x00;
		part.sector = 0x00;
		part.cyl = 0x00;
		/* ending CHS address */
		part.end_head = 0x00;
		part.end_sector = 0x00;
		part.end_cyl = 0x00;
		#endif

		if (!i) { /* make base MBR with extended partition */

			/* partitions of movinand have not boot code */
			part.boot_ind = 0x00;

			part.sys_ind = MS_EX_PART;

			start_sector = MBR_RESERVED_SECTORS;
			memcpy(part.start4, &start_sector, sizeof(start_sector));
			sector_size = movi_total_sectors - MBR_RESERVED_SECTORS;
			memcpy(part.size4, &sector_size, sizeof(start_sector));

			if (DEBUG) {
				printf("== Base MBR with extended partition ==\n");
				printf("start sector : 0x%x \n", start_sector);
				printf("sector_size  : 0x%x \n", sector_size);
				printf("======================================\n");
			}

			memcpy(mbr_image+offset, &part, sizeof(part));
			offset += sizeof(part);

			/* signature of MBR */
			mbr_image[510] = 0x55;
			mbr_image[511] = 0xaa;

			/* write MBR to temporary file */
			if ((written = write(output_fd, mbr_image, MBR_RESERVED_SIZE)) < 0) {
				perror("write");
				ret = FAIL;
				goto mbr_out;
			}

			next_start_sector = MBR_RESERVED_SECTORS;

			/* init to reuse */
			offset = PARTITION_TABLE_OFFSET;
			bzero(mbr_image, MBR_RESERVED_SIZE);
			bzero(&part, sizeof(part));
		}

		/* set current partition information in the MBR of the current partition */
		part.boot_ind = 0x00;
		part.sys_ind = LINUX_NATIVE;

		start_sector = MBR_RESERVED_SECTORS;
		memcpy(part.start4, &start_sector, sizeof(start_sector));
		sector_size = pt[i].partsize >> DEFAULT_SECTOR_BITS;
		memcpy(part.size4, &sector_size, sizeof(start_sector));
		printf("== MBR of partition(%d)\n", i);
		printf("start sector : 0x%x \n", start_sector);
		printf("  part_size  : 0x%x \n", pt[i].partsize);
		printf("sector_size  : 0x%x \n", sector_size);

		memcpy(mbr_image+offset, &part, sizeof(part));
		offset += sizeof(part);

		/* set next partition information in the MBR of the current partition */
		if (i+1 < partno) {
			bzero(&part, sizeof(part));

			part.boot_ind = 0x00;
			part.sys_ind = MS_EX_PART;

			start_sector = next_start_sector + sector_size;
			memcpy(part.start4, &start_sector, sizeof(start_sector));

			/* read next partition information & make sector size of next partition */
			sector_size = (pt[i+1].partsize >> DEFAULT_SECTOR_BITS) + MBR_RESERVED_SECTORS;
			memcpy(part.size4, &sector_size, sizeof(sector_size));
			printf("start sector : 0x%x \n", start_sector);
			printf("sector_size  : 0x%x \n", sector_size);
			printf("=================================\n");

			memcpy(mbr_image+offset, &part, sizeof(part));

			next_start_sector = start_sector + MBR_RESERVED_SECTORS;
		}

		/* signature of MBR */
		mbr_image[510] = 0x55;
		mbr_image[511] = 0xaa;

		/* write MBR to temporary file */
		if ((written = write(output_fd, mbr_image, MBR_RESERVED_SIZE)) < 0) {
			perror("write");
			ret = FAIL;
			goto mbr_out;
		}

		file_size += MBR_RESERVED_SIZE;

		/* write data of the partition to temporary file */
		if (strcmp(pt[i].szFileName, "\0")) { /* if there is a file of the partition */
			if (DEBUG) {
				printf("open : %s\n", pt[i].szFileName);
			}
			if ((input_fd = open(pt[i].szFileName, O_RDONLY)) < 0) {
				perror("open");
				ret = FAIL;
				goto mbr_out;
			}

			while(1) {
				int n = 0;
				if((n = read(input_fd, buffer, sizeof(buffer))) < 0) {
					perror("read");
					ret = FAIL;
					goto mbr_out;
				}

				if(!n) { /* check end of file */
					printf("file_size : %u\n", file_size);

					/* add padding for sector size(512 bytes) alignment */
					if (file_size % 512) {
						n = 512 - (file_size % 512);
						bzero(buffer, sizeof(buffer));
						if ((written = write(output_fd, buffer, n)) < 0) {
							perror("write");
							ret = FAIL;
							goto mbr_out;
						}
						file_size += written;
						printf("file_size with padding data : %u\n", file_size);
					}

					break;
				}

				if((written = write(output_fd, buffer, n)) < 0) {
					perror("write");
					ret = FAIL;
					goto mbr_out;
				}

				file_size += written;
			}

		}
	}

mbr_out:

	free(mbr_image);

	if (output_fd >= 0)
		close(output_fd);

	if (input_fd >= 0)
		close(input_fd);

	return ret;
}

static unsigned int read_pit(const char *pit_name, struct partition_table *pt)
{
	struct partition_table *p;

	partition_header hd;
	partition_info info;

	int fd, i;
	int r;

	fd = open(pit_name, O_RDONLY);

	if (fd < 0) {
		fprintf(stderr, "%s: can't open the file %s.\n", __func__, pit_name);
		return 0;
	}

	r = read(fd, &hd, sizeof(hd));
	if (r < 0) {
		fprintf(stderr, "%s: can't read the file %s.\n", __func__, pit_name);
		return 0;
	}

	printf("%s: PIT magic = 0x%x\n", __func__, hd.dwMagic);
	printf("%s: partition count = %d\n", __func__, hd.nCount);

	for (i = 0; i < hd.nCount; i++) {
		r = read(fd, &info, sizeof(info));
		if (r < 0) {
			fprintf(stderr, "%s: can't read the file %s.\n", __func__, pit_name);
			return 0;
		}

		if (info.nID == 0x0)
			continue;

		printf("=======================================\n");
		printf("partition id      =   %d\n", info.nID);
		printf("partition BlkSize =   %d\n", info.dwBlkSize);
		printf("partition BlkLen  =   %d\n", info.dwBlkLen);
		printf("partition size    =   %u\n", (unsigned int)(info.dwBlkLen * (info.dwBlkSize * 1024)));
		printf("file name         =   %s\n", info.szFileName);

		p = (pt + i);

		p->partno = info.nID;
		p->attr = 0x1002;	/* only RO */
		p->partsize = info.dwBlkLen * (info.dwBlkSize * 1024);

		strcpy(p->szFileName, info.szFileName);
	}

	return hd.nCount;
}

int main(int argc, char **argv)
{
	int fd = -1;
	int written;
	off64_t offset = 0;
	char temp_buf[256];
	unsigned int partno = 0, i, file_size;
	char const *outfile = NULL;
	PARTITION_HEADER *partition_header = NULL;
	struct partition_table part_table[PARTITION_NUM];
	unsigned int movi_total_sectors;
	int ret = SUCCESS;

	if (argc != 4) {
		printf("Usage: %s [Movinand Size] [PIT file] [Target IMG]\n", argv[0]);
		printf("      unit of [Movinand Size] is Kbytes.\n");
		printf("version : %s\n", VERSION_STRING);
		ret = FAIL;
		goto out;
	}

	movi_total_sectors = atoll(argv[1]) << 1; /* argv[1] * 1024 / SECTOR_SIZE(512) */
	printf("movinand total sectors : %u\n", movi_total_sectors);

	outfile = argv[optind + argc - 2];

	if ((fd = open(outfile, O_WRONLY | O_CREAT | O_TRUNC | O_LARGEFILE, 0666)) < 0) {
		perror("open");
		ret = FAIL;
		goto out;
	}

	/* Make Partition Header : Partiton header size == 512 Byte */

	partition_header = malloc(sizeof(PARTITION_HEADER));
	if (partition_header == NULL)
	{
		fprintf(stderr, "Unable to allocate memory.\n");
		goto out;
	}

	memset(partition_header, 0x00, sizeof(PARTITION_HEADER));
	memcpy(partition_header, "FATHEAD", 7);
	offset += sizeof(PARTITION_HEADER);
	printf("partition header size = %lld\n", (long long) offset);

	if (lseek(fd, offset, SEEK_SET) < 0) {
		perror("lseek");
		ret = FAIL;
		goto out;
	}

	/* Make Partition Information */
	/* read partition info from PIT file */
	partno = read_pit(argv[2], part_table);

	if (DEBUG) {
		int ii = 0;


		for (ii=0;ii<partno;ii++) {
			printf("------------------------------------------\n");
			printf("partno          %d\n", part_table[ii].partno);
			printf("attr          0x%x\n", part_table[ii].attr);
			printf("partsize        %d\n", part_table[ii].partsize);
			printf("FileName        %s\n", part_table[ii].szFileName);
		}

		printf("\n");
	}

	printf("movi_total_sectors    %d \n", movi_total_sectors);

	/* Make each partition image with MBR(Master Boot Record) */
	if (make_image_with_MBR(partition_header, movi_total_sectors, part_table, partno) < 0) {
		printf("make_MBR() error\n");
		ret = FAIL;
		goto out;
	}

	/* Copy data of each partition to single image file */
	printf("===============================\n");
	printf("   Make Singe Movinand Image   \n");
	printf("===============================\n");

	for (i = 0; i < partno; i++) {
		sprintf(temp_buf, "%d.mbr", i);
		if((file_size = make_single_image(fd, temp_buf, offset)) < 0) {
			printf("make_single_image() error\n");
			ret = FAIL;
			goto out;
		}
		memcpy((unsigned char *)partition_header + PART_INFO_BASE_OFFSET + i * PART_INFO_EACH_PART_OFFSET,
				&file_size, sizeof(file_size));

		offset = offset + file_size;
	}
	printf("Movinand image size : %lld bytes\n", (long long)offset);

	/* Write PARTITION_HEADER to image file	 */
	offset = 0;
	if (lseek(fd, offset, SEEK_SET) < 0) {
		perror("lseek");
		ret = FAIL;
		goto out;
	}

	if (DEBUG) {
		int iii;
		unsigned char *iiibuf = (unsigned char *)partition_header;

		for (iii=1; iii<=512;iii++) {
			printf("%02x ", iiibuf[iii-1]);
			if (!(iii%16))
				printf("\n");
		}
	}

	written = write(fd, partition_header, sizeof(PARTITION_HEADER));
	if (written < 0) {
		perror("write failed");
		ret = FAIL;
		goto out;
	}

out:

	if (partition_header)
		free(partition_header);

	if (fd >= 0)
		close(fd);

	if (DEBUG) {
		for(i = 0; i < partno; i++) {
			sprintf(temp_buf, "%d.mbr", i);
			remove(temp_buf);
		}
	}

	if (ret) {
		printf("====================================\n");
		printf("Success. Image : %s\n", outfile);
		printf("====================================\n");
	} else {
		printf("===================\n");
		printf("Fail. Retry please.\n");
		printf("===================\n");

		remove(outfile);
	}

	return 1;
}

