/* xmlgetent: Dump one attribute from xml-pit description file */

#include <libxml/tree.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "partition.h"
#include "xml.h"

char *progname;

char *entry_int[] = { "BinType", "DevType", "ID", "Attribute", "UpdateAttr", "BlockSize",
		      "BlockLength", "Offset", "FileSize", };
char *entry_str[] = { "Name", "FileName", "DeltaName", };

#define NELEMS(array) (sizeof(array)/sizeof(array[0]))


int print_elem(xmlElementPtr partition, const char *elem)
{
     char buf[32];
     int i;

     if (!partition)
	  return -1;

     for (i = 0; i < NELEMS(entry_str); i++)
	  if (!strcmp(entry_str[i], elem)) {
	       get_val_str(partition, elem, buf, sizeof(buf));
	       printf("%s\n", buf);
	       return 0;
	  }

     for (i = 0; i < NELEMS(entry_int); i++)
	  if (!strcmp(entry_int[i], elem)) {
	       printf("%d\n", get_val_int(partition, elem));
	       return 0;
	  }

     return -1;
}

int print_all(xmlElementPtr partition)
{
     char buf[32];
     int i;

     if (!partition)
	  return -1;

     for (i = 0; i < NELEMS(entry_str); i++) {
	  get_val_str(partition, entry_str[i], buf, sizeof(buf));
	  printf("%s %s\n", entry_str[i], buf);
     }

     for (i = 0; i < NELEMS(entry_int); i++)
	  printf("%s %d\n", entry_int[i], get_val_int(partition, entry_int[i]));

     return 0;
}

int xmlgetent(int fd, const char *partname, const char *entry)
{
     xmlDocPtr doc;
     xmlNodePtr root_node, node;
     char attr[32];
     int ret = -1;

     doc = xmlReadFd(fd, NULL, NULL, 0);

     if (!doc) {
	  fprintf(stderr, "failed to parse input stream\n");
	  return -1;
     }

     root_node = xmlDocGetRootElement(doc);

     if (strcmp((const char*)(root_node->name), "root")) {
	  fprintf(stderr, "invalid root element name\n");
	  return -1;
     }

     for (node = root_node->children; node; node = node->next) {
	  if (node->type == XML_ELEMENT_NODE && !strcmp((const char*)(node->name), "partition")) {
	       xmlElementPtr elem = (xmlElementPtr)node;
	       get_val_str(elem, "Name", attr, sizeof(attr));

	       if (!partname)
		    printf("%s\n", attr);
	       else if (!strcmp(attr, partname)) {
		    ret = !entry ? print_all(elem) : print_elem(elem, entry);
		    break;
	       }
	  }
     }

     xmlFreeDoc(doc);
     xmlCleanupParser();

     return ret;
}

void usage(void)
{
     int i;

     printf("usage: %s [-i INPUT_FILE] [-e ENTRY] [PARTNAME]\n"
	    "Dump attributes from xml-pit description file.\n"
	    "\n"
	    "Lists partition names if PARTNAME isn't specifed.  Prints all entries\n"
	    "when PARTNAME is specified but ENTRY isn't. Otherwise, prints only\n"
	    "specified entry.\n"
	    "\n"
	    "List of known entries:\n", progname);

     printf("  string: ");
     for (i = 0; i < NELEMS(entry_str); i++)
	  printf("%s ", entry_str[i]);

     printf("\n     int: ");
     for (i = 0; i < NELEMS(entry_int); i++)
	  printf("%s ", entry_int[i]);

     putchar('\n');
}

int main(int argc, char **argv)
{
     int c;
     int fd = 0;

     /* things we look for */
     char *partname = NULL;      /* partition id */
     char *entry = NULL;         /* entry inside partition */

     progname = argv[0];

     while ((c = getopt(argc, argv, "i:e:")) != -1) {
	  switch (c) {
	  case 'i':
	       fd = open(optarg, O_RDONLY);
	       if (fd < 0) {
		    perror("Unable to open input file");
		    return -1;
	       }
	       break;

	  case 'e':
	       entry = strdup(optarg);
	       if (!entry) {
		    perror("Out of memory");
		    return -1;
	       }
	       break;

	  default:
	       usage();
	       return -1;
	  }
     }

     argv += optind;
     argc -= optind;

     partname = *argv; /* might be null, we handle that in xmlgetent() */

     return xmlgetent(fd, partname, entry);
}
